#ifndef __CATALINA_GRAPHICS_H
#define __CATALINA_GRAPHICS_H

// this is just a convenience so we can use the symbol DOUBLE_BUFFER
// in our call to g_setup, or in the calls to the CGI functions
#ifdef __CATALINA_DOUBLE_BUFFER
#define DOUBLE_BUFFER 1 // no double buffer
#else
#define DOUBLE_BUFFER 0 // use double buffering
#endif

// If the size of the font data ever changes, so must this constant:
//
#define FONT_WORDS 287

// plugin codes (used in requests to plugin cog):
//
#define CGI_setup     1
#define CGI_color     2
#define CGI_width     3
#define CGI_plot      4
#define CGI_line      5
#define CGI_arc       6
#define CGI_vec       7
#define CGI_vecarc    8
#define CGI_pix       9
#define CGI_pixarc   10
#define CGI_text     11
#define CGI_textarc  12
#define CGI_textmode 13
#define CGI_fill     14
#define CGI_loop     15

// We need this type so that all variables are allocated on the stack, 
// even when we are using the XMM memory model. We pass a pointer to 
// this structure to all the functions that need to access the variables.
//
typedef struct {
   long X_TILES;           
   long Y_TILES;           
   long BITMAP_BASE;       
   long BITMAP_SIZE;       
   long PIXEL_WIDTH;       
   long TEXT_XS;           
   long TEXT_YS;           
   long TEXT_SP;           
   long TEXT_JUST;         
   long SLICES[8];         
   long COLORS[4];         
   long ARGS[8];           
   short BASES[32];        
#ifndef __CATALINA_TINY
   short FONT[FONT_WORDS]; // only required when using XMM
#endif   
} g_var;


/* private functions */

void g_fill(g_var *gv, int x, int y, int da, int db, int db2, 
            int linechange, int lines_minus_1);

void g_justify(g_var *gv, void *string_ptr, int *just_x, int *just_y);


/* public functions */

// Get x_tiles
//
int cgi_x_tiles();


// Get y_tiles
//
int cgi_y_tiles();


// Get address of underlying CGI display.
//
// Note that this is always a Hub RAM address.
// The bitmap data will be:
//   (x_tiles * y_tiles) tiles, or
//   ((x_tiles * y_tiles) * 16 * 16 * 2) / 8) bytes.
//
void *cgi_display_base();


// Get address of underlying CGI bitmap to draw on.
//
// We must tell this function if we are double buffering.
//
// Note that this is always a Hub RAM address.
// The bitmap data will be:
//   (x_tiles * y_tiles) tiles, or
//   ((x_tiles * y_tiles) * 16 * 16 * 2) / 8) bytes.
//
void *cgi_bitmap_base(int double_buffer); 


// Get address of underlying CGI screen data.
//
// We must tell this function if we are double buffering.
//
// Note that this is always a Hub RAM address.
// You must provide x and y size.
// The screen data will be (x_tiles * y_tiles) words.
//
void *cgi_screen_data(int double_buffer);


// Get address of underlying CGI color data.
//
// We must tell this function if we are double buffering.
//
// Note that this is always a Hub RAM address.
// You must provide x and y size.
// The colors data will always be 64 longs
//
void *cgi_color_data(int double_buffer);


// Set bitmap parameters
//
//   x_origin       - relative-x center pixel
//   y_origin       - relative-y center pixel
//   double_buffer  - true if double buffering
//
// NOTE: you do not need to provide x_tiles, y_tiles, or the
//       display pointer - this function gets them automatically.
//
void g_setup(g_var *gv, int x_org, int y_org, int double_buffer);
   

// Clear ether the display (if not double buffering),
// or the double buffer bitmap (if double buffering)
//
void g_clear(g_var *gv);


// Copy either the specified bitmap, or (if NULL)
// the double buffer bitmap to the display
// (use for flicker-free display)
//
void g_copy(g_var *gv, void *bitmap_base);


// Set pixel color to two-bit pattern
//
//   color       - color code in bits[1..0]
//
void g_color(g_var *gv, int color); 


// Set pixel width
// actual width is w[3..0] + 1
//
//   width       - 0..15 for round pixels, 16..31 for square pixels
//
void g_width(g_var *gv, int width);


// Set pixel color and width
//
void g_colorwidth(g_var *gv, int color, int width);


// Plot point
//
//   x,y         - endpoint
//
void g_plot(g_var *gv, int x, int y);


// Draw a line to point
//
//   x,y         - endpoint
//
void g_line(g_var *gv, int x, int y);


// Draw an arc
//
//   x,y        - center of arc
//   xr,yr      - radii of arc
//   angle      - initial angle in bits[12..0] (0..$1FFF = 0..359.956)
//   anglestep  - angle step in bits[12..0]
//   steps      - number of steps (0 just leaves (x,y) at initial arc position)
//   arcmode    - 0: plot point(s)
//                1: line to point(s)
//                2: line between points
//                3: line from point(s) to center
//
void g_arc(g_var *gv, int x, int y, int xr, int yr, 
           int angle, int anglestep, int steps, int arcmode);


// Draw a vector sprite
//
//   x,y         - center of vector sprite
//   vecscale    - scale of vector sprite ($100 = 1x)
//   vecangle    - rotation angle of vector sprite in bits[12..0]
//   vecdef_ptr  - address of vector sprite definition
//
//
// Vector sprite definition:
//
//   word $8000|$4000+angle  'vector mode + 13-bit angle 
//                           '(mode: $4000=plot, $8000=line)
//   word length             'vector length
//   ...                     'more vectors
//   ...
//   word 0                  'end of definition
//
void g_vec(g_var *gv, int x, int y, int vecscale, int vecangle, 
           void * vecdef_ptr);


// Draw a vector sprite at an arc position
//
//   x,y         - center of arc
//   xr,yr       - radii of arc
//   angle       - angle in bits[12..0] (0..$1FFF = 0..359.956)
//   vecscale    - scale of vector sprite ($100 = 1x)
//   vecangle    - rotation angle of vector sprite in bits[12..0]
//   vecdef_ptr  - address of vector sprite definition
//
void g_vecarc(g_var *gv, int x, int y, int xr, int yr, int angle, 
              int vecscale, int vecangle, void * vecdef_ptr);
      

// Draw a pixel sprite
//
//   x,y         - center of vector sprite
//   pixrot      - 0: 0, 1: 90, 2: 180, 3: 270, +4: mirror
//   pixdef_ptr  - address of pixel sprite definition
//
//
// Pixel sprite definition:
//
//    word    'word align, express dimensions and center, define pixels
//    byte    xwords, ywords, xorigin, yorigin
//    word    %%xxxxxxxx,%%xxxxxxxx
//    word    %%xxxxxxxx,%%xxxxxxxx
//    word    %%xxxxxxxx,%%xxxxxxxx
//    ...
//
void g_pix(g_var *gv, int x, int y, int pixrot, void *pixdef_ptr);


// Draw a pixel sprite at an arc position
//
//   x,y         - center of arc
//   xr,yr       - radii of arc
//   angle       - angle in bits[12..0] (0..$1FFF = 0..359.956)
//   pixrot      - 0: 0, 1: 90, 2: 180, 3: 270, +4: mirror
//   pixdef_ptr  - address of pixel sprite definition
//
void g_pixarc(g_var *gv, int x, int y, int xr, int yr, int angle, 
              int pixrot, void *pixdef_ptr);


// Draw text
//
//   x,y         - text position (see textmode for sizing and justification)
//   string_ptr  - address of zero-terminated string (it may be necessary to 
//                 call finish immediately afterwards to prevent subsequent 
//                 code from clobbering the string as it is being drawn
void g_text(g_var *gv, int x, int y, void *string_ptr);


// Draw text at an arc position
//
//   x,y         - center of arc
//   xr,yr       - radii of arc
//   angle       - angle in bits[12..0] (0..$1FFF = 0..359.956)
//   string_ptr  - address of zero-terminated string (it may be necessary to 
//                 call finish immediately afterwards to prevent subsequent 
//                 code from clobbering the string as it is being drawn
//
void g_textarc(g_var *gv, int x, int y, int xr, int yr, 
               int angle, void *string_ptr);


// Set text size and justification
//
//   x_scale        - x character scale, should be 1+
//   y_scale        - y character scale, should be 1+
//   spacing        - character spacing, 6 is normal
//   justification  - bits[1..0]: 0..3 = left, center, right, left
//                    bits[3..2]: 0..3 = bottom, center, top, bottom
//
void g_textmode(g_var *gv, int x_scale, int y_scale, 
                int spacing, int justification);


// Draw a box with round/square corners, according to pixel width
//
//   x,y      - box left, box bottom
//
void g_box(g_var *gv, int x, int y, int box_width, int box_height);


// Draw a solid quadrilateral
// vertices must be ordered clockwise or counter-clockwise
//
void g_quad(g_var *gv, int x1, int y1, int x2, int y2, 
            int x3, int y3, int x4, int y4);


// Draw a solid triangle
//
void g_tri(g_var *gv, int x1, int y1, int x2, int y2, int x3, int y3);


// Wait for any current graphics command to finish
// use this to insure that it is safe to manually manipulate the bitmap
//
void g_finish();

#endif
